package lsgui;

public final class OSSupport {
    
    /** OS identifier */
    private static int osType;
    
    /** Is System.nanoTime() supported? */
    private static boolean nanoTimeCapable;
    
    /** Unknown system */
    public static final int OS_UNKNOWN = 0;

        /** Other Windows systems */
    public static final int OS_WINDOWS_OTHER = 1;

    /** Windows XP (NT 5.1) */
    public static final int OS_WINDOWS_XP = 2;
    
    /** Windows Server 2003 (NT 5.2) */
    public static final int OS_WINDOWS_SERVER_2003 = 3;

    /** Windows Vista (NT 6.0) */
    public static final int OS_WINDOWS_VISTA = 4;

    /** Windows 7 (NT 6.1) */
    public static final int OS_WINDOWS_7 = 5;

    /** Linux system */
    public static final int OS_LINUX = 10;
    
    /** Solaris system */
    public static final int OS_SOLARIS = 11;
    
    /** Mac OS system */
    public static final int OS_MAC_OS = 12;
    
    /** All other systems */
    public static final int OS_OTHER = 13;
    
    static {
        osType = OS_UNKNOWN;
        detectOS();
    }
    
    /** No instance allowed */
    private OSSupport() {}
    
    private static void detectOS() {
        String osName = System.getProperty("os.name").toLowerCase();
        if (osName.contains("windows")) {
            String version = System.getProperty("os.version");
            if (version.equals("5.1")) {
                osType = OS_WINDOWS_XP;
            } else if (version.equals("5.2")) {
                osType = OS_WINDOWS_SERVER_2003;
            } else if (version.equals("6.0")) {
                osType = OS_WINDOWS_VISTA;
            } else if (version.equals("6.1")) {
                osType = OS_WINDOWS_7;
            } else {
                osType = OS_WINDOWS_OTHER;
            }
        } else if (osName.contains("linux")) {
            osType = OS_LINUX;
        } else if (osName.contains("solaris")) {
            osType = OS_SOLARIS;
        } else if (osName.contains("mac os")) {
            osType = OS_MAC_OS;
        } else {
            osType = OS_OTHER;
        }
        
        long start;
        long stop;
        long total = 0L;
        int nRepeat = 1000;
        
        for (int i = 0; i < nRepeat; i++) {
            start = System.nanoTime();
            stop = System.nanoTime();
            while (stop == start) {
                stop = System.nanoTime();
            }
            total += (stop - start);
        }
        
        nanoTimeCapable = (total / (double) nRepeat / 1000000.0) < 1.0;
    }
    
    public static int getOSType() {
        return osType;
    }
    
    public static boolean isNanoTimeCapable() {
        return nanoTimeCapable;
    }

    public static boolean isWindows() {
        return osType >= OS_WINDOWS_OTHER && osType <= OS_WINDOWS_7;
    }
    
    public static boolean isWindows6orHigher() {
        return isWindows() && osType >= OS_WINDOWS_VISTA;
    }

    public static String getUserLoginName() {
        return System.getProperty("user.name");
    }
    
    public static String getUserHomeDirectory(String directoryName) {
        String homeDir = System.getProperty("user.home");
        String fileSep = System.getProperty("file.separator");
        homeDir += fileSep;
        if (isWindows6orHigher()) {
            homeDir += "AppData" + fileSep + "LocalLow" + fileSep
                    + directoryName + fileSep;
        } else {
            homeDir += "." + directoryName + fileSep;
        }
        return homeDir;
    }
    
    public static boolean createUserHomeDirectory(String directoryName) {
        String path = getUserHomeDirectory(directoryName);
        java.io.File dirPath = new java.io.File(path);
        if (!dirPath.exists()) {
            return dirPath.mkdirs();
        }
        return true;
    }
    
}
