package lsgui.jna.data;

import java.awt.image.BufferedImage;
import java.io.IOException;
import lsgui.util.LaterationImage;
import lsgui.util.LaterationImage.RenderedImage;

/**
 * Targa image reader and writer for LS².
 */
public class ShooterTGAImage {
    
    private int offset;
    private byte[] buffer;

    private static final int TGA_HEADER_SIZE = 38;
    private static final int TGA_COLOR_MAP_SIZE = 768;
    
    /** Error color index in color table (also anchor color) */
    private static final byte COLOR_ERROR = (byte)255;
    
    private static final float CALC_EPSILON = 0.01f;
    public static final float COMPARE_EPSILON = 16;
    public static final float VERY_GOOD_VALUES_COLOR = 100.0f / 2.0f;
    
    /**
     * Used for displaying results of algorithms and LBEs.
     */
    public static final int IMAGE_MODE_ONE = 1;
    
    /**
     * Used for displaying results of comparisons between algorithms or LBEs.
     */
    public static final int IMAGE_MODE_TWO = 2;
    
    /**
     * Used for displaying results of comparisons between algorithms and LBEs.
     */
    public static final int IMAGE_MODE_THREE = 3;
    
    public ShooterTGAImage(LaterationImage image, int imageMode) {
        // make copy of data and leave original data unchanged
        LaterationImage.LaterationImageData data = image.getLaterationImageData();
        float[] result = new float[data.getImageData().length];
        System.arraycopy(data.getImageData(), 0, result, 0, result.length);
        // build color map and prepare image data
        this.offset = 0;
        buffer = new byte[TGA_HEADER_SIZE + TGA_COLOR_MAP_SIZE + result.length];
        writeHeader(buffer, image.getImageWidth(), image.getImageHeight());
        switch (imageMode) {
            case IMAGE_MODE_ONE:
                writeColorMap(buffer, TGA_HEADER_SIZE);
                cropAll(result, 250);
                break;
            case IMAGE_MODE_TWO:
                writeColorMapForCompare(buffer, TGA_HEADER_SIZE);
                clampAndAddAll(result, -127.0f, 127.0f, 127.0f);
                break;
            case IMAGE_MODE_THREE:
                writeColorMapForLowerBoundCompare(buffer, TGA_HEADER_SIZE);
                // add small EPSILON value to avoid very small negative
                // values close to zero because of floating point arithmetic
                clampAndAddAll(result, -1.0f, 253.0f, 1.0f + CALC_EPSILON);
                break;
            default:
                throw new IllegalArgumentException("illegal image mode!");
        }
        prepareImageData(result);
    }

    public static int btoi(byte b) {
        int a = b;
        return (a < 0 ? 256 + a : a);
    }
    
    public static byte[] buildColorMap(double VERY_GOOD_VALUES_COLOR) {
        int i = 0;
        byte[] map = new byte[768];
        // build color table
        do {
            if (i <= (VERY_GOOD_VALUES_COLOR)) {
                tgaSetColor(map, i, 0, 5*i, 255, 5*i);
            } else {
                int color = (int)(250 - (i-50)*(250.0/200.0));
                tgaSetColor(map, i, 0, color, color, color);
            }
            i++;
        } while (i < 250);

        // Max failure in blue
        tgaSetColor(map, 250, 0, 0, 0, 255);
        // Not defined, set to blue
        tgaSetColor(map, 251, 0, 0, 0, 255);
        tgaSetColor(map, 252, 0, 0, 0, 255);
        tgaSetColor(map, 253, 0, 0, 0, 255);
        tgaSetColor(map, 254, 0, 0, 0, 255);
        // Anchors in red
        tgaSetColor(map, 255, 0, 255, 0, 0);
        return map;
    }
    
    private static void tgaSetColor(byte[] colorMap, int i, int offset, int r, int g, int b) {
        colorMap[offset + i * 3 + 2] = (byte)r;
        colorMap[offset + i * 3 + 1] = (byte)g;
        colorMap[offset + i * 3 + 0] = (byte)b;
    }

    /** First 38 bytes are the header */
    private static void writeHeader(byte[] header, int width, int height) {
        header[0] = 20;                             // size of ID field that follows 18 byte header (0 usually)
        header[1] = 1;                              // type of colour map 0=none, 1=has palette
        header[2] = 1;                              // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed
        header[3] = -1;                             // first colour map entry in palette
        header[4] = -1;                             // first colour map entry in palette
        header[5] = -1;                             // number of colours in palette
        header[6] = -1;                             // number of colours in palette
        header[7] = 24;                             // number of bits per palette entry 15,16,24,32
        header[8] = 0;                              // image x origin
        header[9] = 0;                              // image x origin
        header[10] = 0;                             // image y origin
        header[11] = 0;                             // image y origin
        header[12] = (byte)(width & 0x00ff);        // image width in pixels
        header[13] = (byte)((width & 0xff00) >> 8); // image width in pixels
        header[14] = (byte)(height & 0x00ff);       // image height in pixels
        header[15] = (byte)((height & 0xff00) >> 8);// image height in pixels
        header[16] = 8;                             // image bits per pixel
        header[17] = 32;                            // image descriptor bits
    }
    
    private static void writeColorMap(byte[] map, int offset) {
        int i = 0;
        // build color table
        do {
            if (i <= VERY_GOOD_VALUES_COLOR) {
                tgaSetColor(map, i, offset, 5 * i, 255, 5 * i);
            } else {
                int color = (int)(250 - (i-50) * (250.0 / 200.0));
                tgaSetColor(map, i, offset, color, color, color);
            }
            i++;
        } while (i < 250);

        // Max failure in blue
        tgaSetColor(map, 250, offset, 0, 0, 255);
        // Not defined, set to blue
        tgaSetColor(map, 251, offset, 0, 0, 255);
        tgaSetColor(map, 252, offset, 0, 0, 255);
        tgaSetColor(map, 253, offset, 0, 0, 255);
        tgaSetColor(map, 254, offset, 0, 0, 255);
        // Anchors in red
        tgaSetColor(map, 255, offset, 255, 0, 0);
    }
    
    private static void writeColorMapForCompare(byte[] map, int offset) {
        for (int i = 0; i < 255; i++) {
            if (i < 127 - COMPARE_EPSILON / 2) {
                // green - algorithm 1 is better than algorithm 2
                tgaSetColor(map, i, offset, i, 255, i);
            } else if (127 - COMPARE_EPSILON / 2 <= i && i <= 127 + COMPARE_EPSILON / 2) {
                // grey - algorithms are within the same error range EPSILON
                tgaSetColor(map, i, offset, i, i, i);
            } else if (i > 127 + COMPARE_EPSILON / 2) {
                // blue - algorithm 2 is better than algorithm 1
                tgaSetColor(map, i, offset, 254 - i, 254 - i, 255);
            }
        }
        // anchors in red
        tgaSetColor(map, 255, offset, 255, 0, 0);
    }
    
    private static void writeColorMapForLowerBoundCompare(byte[] map, int offset) {
        // blue - algorithm better than lower bound
        tgaSetColor(map, 0, offset, 0, 0, 255);
        for (int i = 1; i < 255; i++) {
            // grey - algorithm worse than lower bound (from light to dark)
            tgaSetColor(map, i, offset, 254 - i, 254 - i, 254 - i);
        }
        // anchors in red
        tgaSetColor(map, 255, offset, 255, 0, 0);
    }
    
    private void prepareImageData(float[] result) {
        for (int i = 0; i < result.length; i++) {
            if (!LS2Result.isError(result[i])) {
                buffer[TGA_HEADER_SIZE + TGA_COLOR_MAP_SIZE + i] = (byte) result[i];
            } else {
                buffer[TGA_HEADER_SIZE + TGA_COLOR_MAP_SIZE + i] = COLOR_ERROR;
            }
        }
    }
    
    private int read() {
        return btoi(buffer[offset++]);
    }
    
    /**
     * Render TGA image to display in GUI.
     * <p>
     * A buffered image with some statistic information
     * is stored in the given rendered image object.
     * 
     * @param rImage Rendered image to store information.
     * 
     * @throws IOException If an I/O error occurs.
     */
    public void decode(RenderedImage rImage) throws IOException {

        // Skipping header
        for (int i = 0; i < 12; i++) {
            read();
        }

        // Read width and height of image
        int width = read() + (read() << 8);
        int height = read() + (read() << 8);

        // Skipping rest of header
        for (int i = 0; i < 22; i++) {
            read();
        }

        // Reading color table
        int[] colors = new int[256];
        for (int i = 0; i < 256; ++i) {
            int b = read();
            int g = read();
            int r = read();
            colors[i] = 0xff000000 | (r << 16) | (g << 8) | b;
        }

        // Reading data
        int[] pixels = new int[width * height];
        for (int i = 0; i < pixels.length; i++) {
            int nb = read();
            pixels[i] = colors[nb];
            // Account histogram data
            if (nb < 255) {
                rImage.histogram[nb]++;
            }
            if (nb == btoi(COLOR_ERROR)) {
                rImage.positionErrors++;
            }
        }

        // Create buffered image with decoded data
        rImage.image = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
        rImage.image.setRGB(0, 0, width, height, pixels, 0, width);
    }
    
    private static void clampAndAddAll(float[] result, float min, float max, float add) {
        for (int i = 0; i < result.length; i++) {
            if (!LS2Result.isError(result[i])) {
                result[i] = clamp(result[i], min, max) + add;
            }
        }
    }
    
    private static void cropAll(float[] result, float max) {
        for (int i = 0; i < result.length; i++) {
            if (!LS2Result.isError(result[i]) && result[i] >= max) {
                result[i] = max;
            }
        }
    }
    
    private static float clamp(float val, float min, float max) {
        return val < min ? min : val > max ? max : val;
    }
        
}
